@@ -5,7 +5,7 @@ This module has been written as part of the Cherokee project:
                http://www.cherokee-project.com/
 """
 
-# Copyright (c) 2006, Alvaro Lopez Ortega <alvaro@alobbs.com>
+# Copyright (c) 2006-2009, Alvaro Lopez Ortega <alvaro@alobbs.com>
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -39,9 +39,14 @@ import SocketServer
 import traceback
 import socket
 import errno
+import time
+import sys
+import os
 
-__version__ = '1.4'
-__author__  = 'Alvaro Lopez Ortega'
+__version__   = '1.11'
+__author__    = 'Alvaro Lopez Ortega'
+__copyright__ = 'Copyright 2009, Alvaro Lopez Ortega'
+__license__   = 'BSD'
 
 
 class SCGIHandler (SocketServer.StreamRequestHandler):
@@ -50,16 +55,43 @@ class SCGIHandler (SocketServer.StreamRequestHandler):
         self.post   = None
         SocketServer.StreamRequestHandler.__init__ (self, request, client_address, server)
 
-    def __safe_read (self, lenght):
-         while True: 
+    def __safe_read (self, length):
+        info = ''
+        while True:
+            if len(info) >= length:
+                return info
+
+            chunk = None
             try:
-                return self.rfile.read(lenght)
-            except socket.error, (err, strerr):
-                if err == errno.EAGAIN or \
-                   err == errno.EWOULDBLOCK or \
-                   err == errno.EINPROGRESS:
+                to_read = length - len(info)
+                chunk = os.read (self.rfile.fileno(), to_read)
+                if not len(chunk):
+                    return info
+                info += chunk
+            except OSError, e:
+                if e.errno in (errno.EAGAIN, errno.EWOULDBLOCK, errno.EINPROGRESS):
+                    if chunk:
+                        info += chunk
+                        continue
+                    time.sleep(0.01)
+                    continue
+
+    def send(self, buf):
+        pending = len(buf)
+        offset  = 0
+
+        while True:
+            if not pending:
+                return
+
+            try:
+                sent = os.write (self.wfile.fileno(), buf[offset:])
+                pending -= sent
+                offset  += sent
+            except OSError, e:
+                if e.errno in (errno.EAGAIN, errno.EWOULDBLOCK, errno.EINPROGRESS):
+                    time.sleep(0.01)
                     continue
-            raise
 
     def __read_netstring_size (self):
         size = ""
@@ -95,8 +127,12 @@ class SCGIHandler (SocketServer.StreamRequestHandler):
             self.env[items[i]] = items[i+1]
 
     def handle_post (self):
+        if self.post:
+            return
+
         if not self.env.has_key('CONTENT_LENGTH'):
             return
+
         length = int(self.env['CONTENT_LENGTH'])
         self.post = self.__safe_read(length)
 
@@ -106,7 +142,8 @@ class SCGIHandler (SocketServer.StreamRequestHandler):
         try:
             self.handle_request()
         except:
-            traceback.print_exc()  # Print the error
+            if sys.exc_type != SystemExit:
+                traceback.print_exc()  # Print the error
 
         try:
             self.finish()          # Closes wfile and rfile
@@ -114,11 +151,14 @@ class SCGIHandler (SocketServer.StreamRequestHandler):
         except: pass
 
     def handle_request (self):
-        self.wfile.write("Content-Type: text/plain\r\n\r\n")
-        self.wfile.write("handle_request() should be overridden")
+        self.send('Status: 200 OK\r\n')
+        self.send("Content-Type: text/plain\r\n\r\n")
+        self.send("handle_request() should be overridden")
 
 
-class SCGIServer(SocketServer.ThreadingTCPServer):
+# TCP port
+#
+class SCGIServer (SocketServer.ThreadingTCPServer):
     def __init__(self, handler_class=SCGIHandler, host="", port=4000):
         self.allow_reuse_address = True
         SocketServer.ThreadingTCPServer.__init__ (self, (host, port), handler_class)
@@ -128,9 +168,29 @@ class SCGIServerFork (SocketServer.ForkingTCPServer):
         self.allow_reuse_address = True
         SocketServer.ForkingTCPServer.__init__ (self, (host, port), handler_class)
 
+# Unix socket
+#
+class SCGIUnixServer (SocketServer.ThreadingUnixStreamServer):
+    def __init__(self, unix_socket, handler_class=SCGIHandler):
+        self.allow_reuse_address = True
+        SocketServer.ThreadingUnixStreamServer.__init__ (self, unix_socket, handler_class)
+
+class SCGIUnixServerFork (SocketServer.UnixStreamServer):
+    def __init__(self, unix_socket, handler_class=SCGIHandler):
+        self.allow_reuse_address = True
+        SocketServer.UnixStreamServer.__init__ (self, unix_socket, handler_class)
+
 
 def ServerFactory (threading=False, *args, **kargs):
+    unix_socket = kargs.get('unix_socket', None)
+
     if threading:
-        return SCGIServer(*args, **kargs)
+        if unix_socket:
+            return SCGIUnixServer (*args, **kargs)
+        else:
+            return SCGIServer(*args, **kargs)
     else:
-        return SCGIServerFork(*args, **kargs)
+        if unix_socket:
+            return SCGIUnixServerFork(*args, **kargs)
+        else:
+            return SCGIServerFork(*args, **kargs)
